<?php

namespace Drupal\whatlinkshere;

/**
 * @file
 * Link Manager class instance for handling references between content entities.
 */

use Drupal\Component\Utility\Html;
use Drupal\Core\Database\Connection;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Path\AliasManagerInterface;
use Drupal\Core\Url;
use Drush\Log\LogLevel;
use Symfony\Component\DependencyInjection\ContainerInterface;

class LinkManager implements LinkManagerInterface {

  /**
   * Drupal entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Drupal database service.
   *
   * @var \Drupal\Core\Database\Connection
   */
  protected $database;

  /**
   * Drupal entity field manager service.
   *
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * Path alias manager service.
   *
   * @var \Drupal\Core\Path\AliasManagerInterface
   */
  protected $pathAliasManager;

  /**
   * Class constructor.
   */
  public function __construct(EntityTypeManagerInterface $entity_type_manager, Connection $database, EntityFieldManagerInterface $entity_field_manager, AliasManagerInterface $path_alias_manager) {
    $this->entityTypeManager = $entity_type_manager;
    $this->database = $database;
    $this->entityFieldManager = $entity_field_manager;
    $this->pathAliasManager = $path_alias_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container) {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('database'),
      $container->get('entity_field.manager'),
      $container->get('path.alias_manager')
    );
  }

  /**
   * {@inheritdoc}
   */
  public function processEntity(EntityInterface $entity) {
    $fields = $this->entityFieldManager->getFieldDefinitions('node', $entity->bundle());

    $types_of_interest = [
      'entity_reference',
      'text_with_summary',
      'text_long',
      'link',
    ];

    // Array to store entity ids of things that this entity references.
    $reference_values = [];

    foreach ($fields as $field) {
      $type = $field->getType();
      $field_name = $field->getName();
      $field_value = NULL;

      // Skip over any fields we don't believe would contain a reference.
      if (in_array($type, $types_of_interest) == FALSE) {
        continue;
      }

      // Array to capture any extracted nids from any field type.
      $extracted_nids = [];

      // Text fields that may contain link markup.
      if ($type == 'text_with_summary' || $type == 'text_long') {
        $field_value = $entity->get($field->getName())->value;
        // Scan for link elements in this chunk of HTML.
        $dom = Html::load($field_value);
        $link_elements = $dom->getElementsByTagName('a');

        foreach ($link_elements as $link) {
          $href = $link->getAttribute('href');

          if (preg_match('/^http/', $href)) {
            // Skip over absolute or external links.
            continue;
          }
          else {
            // Lookup content by path alias.
            $matches = [];
            preg_match('/node\/(\d+)/', $this->pathAliasManager->getPathByAlias($href), $matches);

            if (!empty($matches[1])) {
              $extracted_nids[] = $matches[1];
            }
          }
        }

        // Dedupe the array and store in the 'field values' variable to use in the query later.
        $field_value = array_unique($extracted_nids);
      }

      // Any link fields.
      if ($type == 'link') {
        $link_field_values = $entity->get($field->getName())->getValue();

        foreach ($link_field_values as $link) {
          if (empty($link['uri'])) {
            continue;
          }

          $url = Url::fromUri($link['uri']);

          if ($url->isExternal() == FALSE) {
            $uri_path = '';

            try {
              $uri_path = $url->getInternalPath();
            }
            catch (\UnexpectedValueException $ex) {
              // TODO: pass params as placeholders rather than string interpolation.
              \Drupal::service('logger.channel.php')->log(LogLevel::WARNING, 'Could not resolve internal path for Url (' . $link['uri'] . '), entity ID: ' . $entity->id());
              // Log the problem (likely a broken link) and continue the wider loop.
              continue 2;
            }

            $matches = [];

            if (preg_match('/node\/(\d+)/', $uri_path, $matches)) {
              $extracted_nids[] = $matches[1];
            }
          }
        }

        $field_value = array_unique($extracted_nids);
      }

      // Entity reference fields (only interested in Node for PoC).
      if ($type == 'entity_reference' && $field->getSetting('target_type') == 'node') {
        $field_value = $entity->get($field->getName())->getValue();

        if (!empty($field_value)) {
          foreach ($field_value as $value) {
            $extracted_nids[] = $value['target_id'];
          }

          $field_value = array_unique($extracted_nids);
        }
      }

      // Cast field value to array for ease of iterating.
      $field_value = (array) $field_value;

      // Bundle the gathered values into an array to pass into our DB layer.
      if (count($field_value) > 0) {
        for ($i = 0; $i < count($field_value); $i++) {
          if (empty($field_value[$i])) {
            continue;
          }

          $reference_values[] = [
            'id' => $entity->id(),
            'reference_id' => $field_value[$i],
            'reference_field' => $field_name,
            'delta' => $i,
          ];
        }
      }
    }

    // Delete existing values.
    $this->database->delete('whatlinkshere')
      ->condition('id', $entity->id())
      ->execute();

    // Insert new values.
    $query = $this->database->insert('whatlinkshere')->fields([
      'id',
      'reference_id',
      'reference_field',
      'delta',
    ]);
    foreach ($reference_values as $record) {
      $query->values($record);
    }

    $query->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function deleteEntity(EntityInterface $entity) {
    $this->database->delete('whatlinkshere')->condition('id', $entity->id())->execute();
  }

  /**
   * {@inheritdoc}
   */
  public function getReferenceContent(EntityInterface $entity, int $num_per_page, int $offset, array $sort_options = []) {
    $query = $this->database->select('node_field_data', 'nfd');
    $query->fields('nfd', ['nid', 'title', 'type']);
    $query->innerJoin('whatlinkshere', 'b', 'nfd.nid = b.id');
    $query->addExpression('GROUP_CONCAT(DISTINCT b.reference_field)', 'reference_fields');
    $query->condition('b.reference_id', $entity->id(), '=');
    $query->groupBy('nfd.nid, nfd.title, nfd.type');
    $query->extend('Drupal\Core\Database\Query\TableSortExtender')->orderByHeader($sort_options);
    $query->execute()->fetchAll();

    // First query excludes pager range; just for total result set size.
    $related_content = [
      'total' => count($query->execute()->fetchAll()),
    ];

    // Re-query with range (pager support and sort support).
    $query->range($offset, $num_per_page);
    $result = $query->execute()->fetchAll();

    foreach ($result as $record) {
      $related_content['rows'][] = [
        'nid' => $record->nid,
        'type' => $this->entityTypeManager->getStorage('node_type')->load($record->type)->label(),
        'title' => $record->title,
        // Pad commas with trailing space to improve readability.
        'reference_fields' => str_replace(',', ', ', $record->reference_fields),
      ];
    }

    return $related_content;
  }

}
